@@ -5,7 +5,7 @@
 ;; Author: Protesilaos Stavrou <info@protesilaos.com>
 ;; Maintainer: Protesilaos Stavrou <info@protesilaos.com>
 ;; URL: https://github.com/protesilaos/fontaine
-;; Version: 2.0.0
+;; Version: 2.1.0
 ;; Package-Requires: ((emacs "29.1"))
 
 ;; This file is NOT part of GNU Emacs.
@@ -50,6 +50,18 @@
   '(normal italic oblique reverse-italic reverse-oblique)
   "List of font slants.")
 
+(defvar fontaine-widths
+  '( ultra-condensed
+     extra-condensed
+     condensed
+     semi-condensed demi-condensed
+     normal medium regular
+     semi-expanded demi-expanded
+     expanded
+     extra-expanded
+     ultra-expanded)
+  "List of font widths.")
+
 (defconst fontaine--weights-widget
   '(choice :tag "Font weight (must be supported by the typeface)"
            (const :tag "Normal" normal)
@@ -77,6 +89,24 @@
            (const :tag "Use fallback value" nil))
   "Widget with font slants for `fontaine-presets'.")
 
+(defconst fontaine--widths-widget
+  '(choice :tag "Font width (must be supported by the typeface)"
+           (const ultra-condensed)
+           (const extra-condensed)
+           (const condensed)
+           (const semi-condensed)
+           (const :tag "Alias for semi-condensed" demi-condensed)
+           (const normal)
+           (const :tag "Alias for normal" medium)
+           (const :tag "Alias for normal" regular)
+           (const semi-expanded)
+           (const :tag "Alias for semi-expanded" demi-expanded)
+           (const expanded)
+           (const extra-expanded)
+           (const ultra-expanded)
+           (const :tag "Use fallback value" nil))
+  "Widget with font weights for `fontaine-presets'.")
+
 (defconst fontaine-faces
   '( default fixed-pitch fixed-pitch-serif variable-pitch
      mode-line-active mode-line-inactive header-line
@@ -90,6 +120,7 @@
    `((const :tag ,(format "%s font family" face) ,(intern (format ":%s-family" face))) string)
    `((const :tag ,(format "%s weight" face) ,(intern (format ":%s-weight" face))) ,fontaine--weights-widget)
    `((const :tag ,(format "%s slant" face) ,(intern (format ":%s-slant" face))) ,fontaine--slants-widget)
+   `((const :tag ,(format "%s width" face) ,(intern (format ":%s-width" face))) ,fontaine--widths-widget)
    `((const :tag ,(format "%s height" face) ,(intern (format ":%s-height" face))) float)))
 
 (defcustom fontaine-presets
@@ -106,61 +137,73 @@
      :default-family "Monospace"
      :default-weight regular
      :default-slant normal
+     :default-width normal
      :default-height 100
 
      :fixed-pitch-family nil
      :fixed-pitch-weight nil
      :fixed-pitch-slant nil
+     :fixed-pitch-width nil
      :fixed-pitch-height 1.0
 
      :fixed-pitch-serif-family nil
      :fixed-pitch-serif-weight nil
      :fixed-pitch-serif-slant nil
+     :fixed-pitch-serif-width nil
      :fixed-pitch-serif-height 1.0
 
      :variable-pitch-family "Sans"
      :variable-pitch-weight nil
      :variable-pitch-slant nil
+     :variable-pitch-width nil
      :variable-pitch-height 1.0
 
      :mode-line-active-family nil
      :mode-line-active-weight nil
      :mode-line-active-slant nil
+     :mode-line-active-width nil
      :mode-line-active-height 1.0
 
      :mode-line-inactive-family nil
      :mode-line-inactive-weight nil
      :mode-line-inactive-slant nil
+     :mode-line-inactive-width nil
      :mode-line-inactive-height 1.0
 
      :header-line-family nil
      :header-line-weight nil
      :header-line-slant nil
+     :header-line-width nil
      :header-line-height 1.0
 
      :line-number-family nil
      :line-number-weight nil
      :line-number-slant nil
+     :line-number-width nil
      :line-number-height 1.0
 
      :tab-bar-family nil
      :tab-bar-weight nil
      :tab-bar-slant nil
+     :tab-bar-width nil
      :tab-bar-height 1.0
 
      :tab-line-family nil
      :tab-line-weight nil
      :tab-line-slant nil
+     :tab-line-width nil
      :tab-line-height 1.0
 
      :bold-family nil
      :bold-slant nil
      :bold-weight bold
+     :bold-width nil
      :bold-height 1.0
 
      :italic-family nil
      :italic-weight nil
      :italic-slant italic
+     :italic-width nil
      :italic-height 1.0
 
      :line-spacing nil))
@@ -184,8 +227,8 @@ For each face, Fontaine reads keywords that describe its font
 family, font weight, font slant, and font height.  The name of
 those keywords is composed from the name of the face plus th
 specifier.  For example, with the `default' face, we have
-`:default-family', `:default-height', `:default-weight', and
-`:default-slant'.
+`:default-family', `:default-height', `:default-weight',
+`:default-slant', and `:default-width'.
 
 The properties in more detail:
 
@@ -196,6 +239,8 @@ The properties in more detail:
 
 - The font slant is an unquoted symbol among `fontaine-slants'.
 
+- The font width is an unquoted symbol among `fontaine-widths'.
+
 - The font height is a floating point (like 1.0) which is
   interpreted as a multiple of the default font height.  An
   exception to this is for the `default' face (i.e. the
@@ -241,6 +286,8 @@ Caveats or further notes:
           (plist :options
                  (((const :tag "Default font family" :default-family) string)
                   ((const :tag "Default weight" :default-weight) ,fontaine--weights-widget)
+                  ((const :tag "Default slant" :default-slant) ,fontaine--slants-widget)
+                  ((const :tag "Default width" :default-width) ,fontaine--widths-widget)
                   ((const :tag "Default height" :default-height) natnum)
 
                   ,@(mapcan
@@ -287,20 +334,22 @@ This is then used to restore the last value with the function
    (frame nil)
    (t 0)))
 
-(defun fontaine--set-face-attributes (face family &optional weight slant height frame)
-  "Set FACE font to FAMILY, with optional WEIGHT, SLANT, HEIGHT, FRAME."
+(defun fontaine--set-face-attributes (face family &optional weight slant height width frame)
+  "Set FACE font to FAMILY, with optional WEIGHT, SLANT, HEIGHT, WIDTH, FRAME."
   (let ((frames (fontaine--frame frame)))
     ;; ;; Read this: <https://debbugs.gnu.org/cgi/bugreport.cgi?bug=45920>
     ;; ;; Hence why the following fails.  Keeping it for posterity...
     ;; (set-face-attribute face nil :family family :weight weight :height height)
-    (when (and (symbolp weight)
-               (not (null weight))
+    (when (and weight
+               (symbolp weight)
                (eq (face-attribute face :weight) weight)
                (stringp family))
       (internal-set-lisp-face-attribute face :family family frames))
-    (when (and (symbolp weight) (not (null weight)))
+    (when (and weight (symbolp weight))
       (internal-set-lisp-face-attribute face :weight weight frames))
-    (when (and (symbolp slant) (not (null slant)))
+    (when (and width (symbolp width))
+      (internal-set-lisp-face-attribute face :width width frames))
+    (when (and slant (symbolp slant))
       (internal-set-lisp-face-attribute face :slant slant frames))
     (when (stringp family)
       (internal-set-lisp-face-attribute face :family family frames))
@@ -342,6 +391,7 @@ If FRAME is nil, apply the effect to all frames."
      (plist-get properties (intern (format ":%s-weight" face)))
      (plist-get properties (intern (format ":%s-slant" face)))
      (plist-get properties (intern (format ":%s-height" face)))
+     (plist-get properties (intern (format ":%s-width" face)))
      frame)))
 
 (defun fontaine--set-faces (preset frame)
@@ -364,23 +414,47 @@ If FRAME is nil, apply the effect to all frames."
                symbol))
            fontaine-presets)))
 
-(defun fontaine--set-fonts-prompt ()
-  "Prompt for font set (used by `fontaine-set-fonts')."
-  (let* ((def (nth 1 fontaine--font-display-hist))
-         (prompt (if def
-                     (format "Apply font configurations from PRESET [%s]: " def)
-                   "Apply font configurations from PRESET: ")))
-    (intern
-     (completing-read
-      prompt
-      (fontaine--presets-no-fallback)
-      nil t nil 'fontaine--font-display-hist def))))
+(defun fontaine--get-preset-symbols ()
+  "Return list of the `car' of each element in `fontain-presets'."
+  (delq nil
+        (mapcar
+         (lambda (element)
+           (when-let ((first (car element))
+                      (_ (not (eq first t))))
+             first))
+         fontaine-presets)))
 
 (defvar fontaine-current-preset nil
   "Current font set in `fontaine-presets'.
 This is the preset last used by `fontaine-set-preset'.  Also see
 the command `fontaine-apply-current-preset'.")
 
+(defun fontaine--get-first-non-current-preset (history)
+  "Return the first element of HISTORY which is not `fontaine-current-preset'.
+Only consider elements that are still part of the `fontaine-presets',
+per `fontaine--get-preset-symbols'."
+  (catch 'first
+    (dolist (element history)
+      (when (stringp element)
+        (setq element (intern element)))
+      (when (and (not (eq element fontaine-current-preset))
+                 (member element (fontaine--get-preset-symbols)))
+        (throw 'first element)))))
+
+(defun fontaine--set-fonts-prompt (&optional prompt)
+  "Prompt for font set (used by `fontaine-set-fonts').
+If optional PROMPT string, use it for the prompt, else use one that asks
+for a preset among `fontaine-presets'."
+  (let* ((def (symbol-name (fontaine--get-first-non-current-preset fontaine--font-display-hist)))
+         (prompt (if prompt
+                     (format-prompt prompt nil)
+                   (format-prompt "Apply font configurations from PRESET" def))))
+    (intern
+     (completing-read
+      prompt
+      (fontaine--presets-no-fallback)
+      nil t nil 'fontaine--font-display-hist def))))
+
 ;;;###autoload
 (defun fontaine-set-preset (preset &optional frame)
   "Set font configurations specified in PRESET.
@@ -436,6 +510,19 @@ which this function ignores"
       (fontaine-set-preset current)
     (user-error "The `fontaine-current-preset' is not among `fontaine-presets'")))
 
+;;;###autoload
+(defun fontaine-toggle-preset ()
+  "Toggle between the last two known Fontaine presets.
+These are the presets that were set with `fontaine-set-preset'.  If
+there are no two selected presets, then prompt the user to set a preset.
+
+As a final step, call the `fontaine-set-preset-hook'."
+  (interactive)
+  (fontaine-set-preset
+   (if-let ((previous (fontaine--get-first-non-current-preset fontaine--font-display-hist)))
+       previous
+     (fontaine--set-fonts-prompt "No previous preset to toggle; select PRESET"))))
+
 ;;;; Store and restore preset
 
 (defvar fontaine--preset-history '()
@@ -443,7 +530,7 @@ which this function ignores"
 
 ;;;###autoload
 (defun fontaine-store-latest-preset ()
-  "Write latest cursor state to `fontaine-latest-state-file'.
+  "Write latest state to `fontaine-latest-state-file'.
 Can be assigned to `kill-emacs-hook'."
   (when-let ((hist fontaine--preset-history)
              (latest (car hist))
